#!/usr/bin/perl -w
#######################################################
# logdog.pl
# Written by: Brandon Zehm <caspian@dotconf.net>
#
# License:
#
# logdog.pl (hereafter referred to as "program") is free software;
#   you can redistribute it and/or modify it under the terms of the GNU General
#   Public License as published by the Free Software Foundation; either version
#   2 of the License, or (at your option) any later version.
# Note that when redistributing modified versions of this source code, you
#   must ensure that this disclaimer and the above coder's names are included
#   VERBATIM in the modified code.
#
# Disclaimer:
# This program is provided with no warranty of any kind, either expressed or
#   implied.  It is the responsibility of the user (you) to fully research and
#   comprehend the usage of this program.  As with any tool, it can be misused,
#   either intentionally (you're a vandal) or unintentionally (you're a moron).
#   THE AUTHOR(S) IS(ARE) NOT RESPONSIBLE FOR ANYTHING YOU DO WITH THIS PROGRAM
#   or anything that happens because of your use (or misuse) of this program,
#   including but not limited to anything you, your lawyers, or anyone else
#   can dream up.  And now, a relevant quote directly from the GPL:
#
#                           NO WARRANTY
#
#  11. BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
# FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN
# OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
# PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED
# OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS
# TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE
# PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,
# REPAIR OR CORRECTION.
#
#   ---
#
# Whee, that was fun, wasn't it?  Now let's all get together and think happy
#   thoughts - and remember, the same class of people who made all the above
#   legal spaghetti necessary are the same ones who are stripping your rights
#   away with DVD CCA, DMCA, stupid software patents, and retarded legal
#   challenges to everything we've come to hold dear about our internet.
#   So enjoy your dwindling "freedom" while you can, because 1984 is coming
#   sooner than you think.  :[
#
#######################################################
#
# Purpose: To create a program that is capable of monitoring messages passed
#          through syslogd, triggering alerts if needed.
#          
#
# Layout:  Version two of logdog takes a different approach to monitoring
#          than the last version.  This version is designed to listen to 
#          a fifo socket for incoming messages.  Syslogd can log to a fifo
#          socket simply by putting a pipe (|) character before the name 
#          of the file.  So to have syslogd send its messages to /dev/fifo
#          you would simply have an entry in your syslog.conf like this:
#          *.*                                     |/dev/fifo
#
# Features:
#       - Monitor syslogd messages for key words and phrases and run system 
#         commands based on content.
#       - HUP signal is intercepted correctly and causes logdog to reload
#         its configuration and refresh all filehandles.
#       - Reads data from syslogd via a FIFO for efficiency and low latency alerts.
#       - Debugging mode with multiple debugging levels (enabled via a 
#         variable at the top of this script or command line arguments)
#          
#
use strict;
use IO::Socket;



## Global Variable(s)
my %conf = (
    "programName"          => $0,                                ## The name of this program
    "version"              => '2.0-RC1',                         ## The version of this program
    "authorName"           => 'Brandon Zehm',                    ## Author's Name
    "authorEmail"          => 'caspian@dotconf.net',             ## Author's Email Address
    "debug"                => 0,                                 ## Default debug level
    "configurationFile"    => '/etc/logdog.conf',                ## Default configuration file
    "fifo"                 => '/var/log/fifo.logdog',            ## Default FIFO to read from (you can create a fifo anywhere by using the mkfifo command)
    
    ## The following entries are used later in the program
    "mode"                 => '',                                ## Used later by the program
    "logging"              => '',                                ## If this is true the printmsg function prints to the log file

    ## The folling entries get their values from the configuration file
    "logFile"              => '',

);
$conf{'programName'} =~ s/(.)*[\/,\\]//;                         ## Remove path from filename
$0 = "[$conf{'programName'}]";


## Variables populated by the configuration file
my %alerts = ();
my %monitors = ();
my @ignore = ();



#############################
##
##      MAIN PROGRAM
##
#############################

## Initialize
initialize();

## Process Command Line
processCommandLine();

## Read Configuration File
readConfigurationFile($conf{'configurationFile'});

## Open log file
if (openLogFile($conf{'logFile'})) { quit("ERROR: Opening the file [$conf{'logFile'}] returned the error: $!", 1); }

## Open fifo
if (openFIFO($conf{'fifo'})) { quit("ERROR: Connecting to fifo [$conf{'fifo'}] returned the error: $!", 1); }







#########################
##                     ##
##  MAIN PROGRAM LOOP  ##
##                     ##
#########################


## Become deamon
if ($conf{"mode"} =~ /daemon/i) {
    
    if ( ($conf{'debug'} >= 1) or (fork() == 0) ) {
        
        ## Close normal output methods since we won't be needing them in daemon mode
        close STDOUT unless ($conf{"debug"} >= 1);
        close STDERR unless ($conf{"debug"} >= 1);
        
        ## Set the mode to running so that messages get logged rather than printed to the screen
        $conf{"mode"} = "running";
        
        ## Setup HUP handler function - this will re-read the configuration file, re-open the fifo, log file, etc
        $SIG{'HUP'}   = sub { handleHUP(); };
        
        
        ## Forever
        while (1) {
            
            ## Read a line from the fifo
            my $message = <FIFO>;
            chomp $message if ($message);
            my $alert = "";
            
            if ($message) {
                ## Check to see if the message matches one of our monitor phrases
                ## If it does then add the approperiate alert command to a list of alerts to run later.
                foreach my $key (keys(%monitors)) { if ($message =~ /$key/i) { $alert .= "$monitors{$key}," unless ($alert =~ /$monitors{$key},/); } }
                
                
                ## Check to see if the message matches one of the ignore entries
                ## If it does, then empty the list of alerts to run
                if ($alert) { foreach my $key (@ignore) { if ($message =~ /$key/i) { $alert = ""; } } }
                
                
                ## Run approperiate alert commands if needed
                if ($alert) {
                    foreach my $key (split(/,/, $alert)) {
                        $key =~ s/^\s+//; $key =~ s/\s+$//;
                        my $command = $alerts{$key};
                        if ($command) {
                            $command =~ s/MESSAGE/$message/g;
                            printmsg("INFO: Running the alert command [$key] with message [$message]",0);
                            system("$command &");
                        }
                        else {
                            printmsg("WARNING: I was about to execute the alert [$key], but it does not appear to have a command associated with it!  Logdog will continue to run, but you should fix and reload the configuration file!");
                        }
                    }
                }
            }
        }
    }
}




## Quit (should never get here)
quit("",0);
































######################################################################
## Function:    help ()
##
## Description: For all those newbies ;) 
##              Prints a help message and exits the program.
## 
######################################################################
sub help {
print <<EOM;

$conf{'programName'}-$conf{'version'} by $conf{'authorName'} <$conf{'authorEmail'}>

Usage:  $conf{'programName'} -d [options]
  
  Required:
    -d,--daemon                   become a daemon
  
  Optional:
    -c <file>,--config=<file>     specify a different config file
    -f <file>,--fifo=<file>       specify a different fifo to listen to
    -v                            verbosity - use multiple times for greater effect
          
EOM
quit("", 1);
}










######################################################################
##  Function: initialize ()
##  
##  Does all the script startup jibberish.
##  
######################################################################
sub initialize {

  ## Set STDOUT to flush immediatly after each print  
  $| = 1;

  ## Intercept signals
  $SIG{'QUIT'}  = sub { quit("$$ - $conf{'programName'} - EXITING: Received SIG$_[0]", 1); };
  $SIG{'INT'}   = sub { quit("$$ - $conf{'programName'} - EXITING: Received SIG$_[0]", 1); };
  $SIG{'KILL'}  = sub { quit("$$ - $conf{'programName'} - EXITING: Received SIG$_[0]", 1); };
  $SIG{'TERM'}  = sub { quit("$$ - $conf{'programName'} - EXITING: Received SIG$_[0]", 1); };
  
  ## ALARM and HUP signals are not supported in Win32
  unless ($^O =~ /win/i) {
      $SIG{'HUP'}   = sub { quit("$$ - $conf{'programName'} - EXITING: Received SIG$_[0]", 1); };
      $SIG{'ALRM'}  = sub { quit("$$ - $conf{'programName'} - EXITING: Received SIG$_[0]", 1); };
  }
  
  return(1);
}









######################################################################
##  Function: processCommandLine ()
##  
##  Processes command line storing important data in global var %conf
##  
######################################################################
sub processCommandLine {
  
  
  ############################
  ##  Process command line  ##
  ############################
  
  my $numargv = @ARGV;
  help() unless ($numargv);
  my $counter = 0;
  for ($counter = 0; $counter < $numargv; $counter++) {

    if ($ARGV[$counter] =~ s/^-d//i) {                   ## Daemon ##
        $conf{'mode'} = "daemon";
    }
    
    elsif ($ARGV[$counter] =~ s/^--config=//i) {         ## Config File ##
        $conf{'configurationFile'} = $';
    }
    
    elsif ($ARGV[$counter] =~ /^-c$/i) {                 ## Config File ##
        $counter++;
        $conf{'configurationFile'} = $ARGV[$counter];
    }
    
    elsif ($ARGV[$counter] =~ s/^--fifo=//i) {           ## FIFO Socket ##
        $conf{'fifo'} = $';
    }
    
    elsif ($ARGV[$counter] =~ /^-f$/i) {                 ## FIFO Socket ##
        $counter++;
        $conf{'fifo'} = $ARGV[$counter];
    }
    
    elsif ($ARGV[$counter] =~ /^-v$/i) {                 ## Verbosity ##
        $conf{'debug'}++;
    }

    elsif ($ARGV[$counter] =~ /^-h$|^--help$/i) {        ## Help ##
        help();
    }
    
    else {
        printmsg("\"$ARGV[$counter]\" is not a recognised option!", 0);
        help();
    }
    
  }
  
  ## Print help if we're not becoming a daemon.
  if (!$conf{'mode'}) {
      printmsg("WARNING: The -d parameter must be passed to become a daemon.",0);
      help();
  }
  
  ###################################################
  ##  Verify required variables are set correctly  ##
  ###################################################
  my @required = (
                    'configurationFile',
                    'fifo',
                    'mode',
  );
  foreach (@required) {
      if (!$conf{$_}) {
          quit("ERROR: Value [$_] was not set after parsing command line arguments!", 1);
      }
  }
  
  return(1);
}



















###############################################################################################
## FUNCTION:    
##   readConfigurationFile ( string $filename )
## 
## 
## DESCRIPTION: 
##   Reads the file $filename, and stores information from the file into 
##   the global array $conf, and a few other global arrays.  This version
##   of this function is specifically modified to support logdog's configuration
##   file.
##
## Example: 
##   readConfigurationFile("/etc/configuration.conf");
##
###############################################################################################
sub readConfigurationFile {
  
  ## Get incoming variables
  my %incoming = ();
  (
    $incoming{'confFile'},
  ) = @_;
  $incoming{'section'} = "";
  
  ## Open the file
  open(CONFFILE, $incoming{'confFile'}) or quit("$$ - ERROR While opening the configuration file [$incoming{'confFile'}].  The error was: $!\n",1);
  while (<CONFFILE>) {  ## FOR EACH LINE
    my $line = $_;
    chomp $line;           ## Remove line ending characters
    $line =~ s/\#.+$//;    ## Remove comments
    $line =~ s/\/\/.+$//;  ## Remove comments
    $line =~ s/^\s+//;     ## Remove leading white space
    $line =~ s/\s+$//;     ## Remove ending white space
    
    ## Get the section
    if ($line =~ /section/) {
      $incoming{'section'} = (split(/:/, $line))[1];
      $incoming{'section'} =~ s/^\s+//;     ## Remove leading white space
      $incoming{'section'} =~ s/\s+$//;     ## Remove ending white space
      $line = "";
    } 
    
    
    if ($line ne "") {
      
      ## Assign data for section: general
      if ($incoming{'section'} =~ /general/i) {
        $conf{(split(/\s+=>\s+/, $line))[0]} = (split(/\s+=>\s+/, $line))[1];
      }
      
      ## Assign data for section: alerts
      if ($incoming{'section'} =~ /alerts/i) {
        $alerts{(split(/\s+=>\s+/, $line))[0]} = (split(/\s+=>\s+/, $line))[1];
      }
      
      ## Assign data for section: monitors
      if ($incoming{'section'} =~ /monitors/i) {
        $monitors{(split(/\s+=>\s+/, $line))[0]} = (split(/\s+=>\s+/, $line))[1];
      }
      
      ## Assign data for section: ignore
      if ($incoming{'section'} =~ /ignore/i) {
        push(@ignore, $line);
      }
      
    }
  }
  close(CONFFILE);
  
  ## DEBUGGING:
  printmsg ("Configuration: ", 3); foreach my $key (keys(%conf))     { printmsg ("Key: $key =>\t Value: $conf{$key}", 3) if ($key and $conf{$key}); }
  printmsg ("Alerts: ", 3);        foreach my $key (keys(%alerts))   { printmsg ("Key: $key\t Value: $alerts{$key}", 3) if ($key and $alerts{$key}); }
  printmsg ("Monitor List: ", 3);  foreach my $key (keys(%monitors)) { printmsg ("Key: $key\t Value: $monitors{$key}", 3) if ($key and $monitors{$key}); }
  printmsg ("Ignore List: ", 3);   foreach my $key (@ignore)         { printmsg ("Key: $key", 3) if ($key); }
  
  
  ## Verify that all alerts have a command associated
  foreach my $key (keys(%alerts))   { if (!$alerts{$key}) { printmsg("WARNING: The alert definition [$key] does not have any command associated with it.   Logdog will continue to run, but you should fix and reload the configuration file!", 0); } }
  
  ## Verify that all monitors have valid alerts
  foreach my $key (keys(%monitors)) { foreach my $alert (split(/,/, $monitors{$key})) { if (!$alerts{$alert}) { printmsg ("WARNING: The monitor [$key] references an alert [$alert] that does not exsist! Logdog will continue to run, but you should fix and reload the configuration file!", 0); } } }
  
  
  return(1);
}


















#########################################################
# FUNCTION: handleHUP()
# 
# Handles HUP signals:
#  - Re-reads the configuration file
#  - Logs a message to the log file
#  - Closes and opens the log file 
#  - Closes and opens the fifo
#
#########################################################
sub handleHUP  {
    
    printmsg("$$ - $conf{'programName'} - Received SIGHUP: Reloading config and refreshing all file handles", 0);
    
    ######################################
    ##  Re-read the configuration file  ##
    ######################################
    readConfigurationFile($conf{'configurationFile'});
    
    
    ###################################
    ##  Close and open the log file  ##
    ###################################
    close LOGFILE;
    if (openLogFile($conf{'logFile'})) { quit("ERROR: Opening the file [$conf{'logFile'}] returned the error: $!", 1); }
    
    
    ###############################
    ##  Close and open the fifo  ##
    ###############################
## This causes the program to segfault unpredictibly, and it doesn't help anything to re-open it anyway.
#    close FIFO;
#    if (openFIFO($conf{'fifo'})) { quit("ERROR: Connecting to fifo [$conf{'fifo'}] returned the error: $!", 1); }
    
    
    ## Go back to the normal routine
    return(1);
}














###############################################################################################
##  Function:    printmsg (string $message, int $level)
##
##  Description: Handles all messages - logging them to a log file, 
##               printing them to the screen or both depending on
##               the $level passed in, $conf{'debug'} and wether
##               $conf{'mode'}.
##
##  Input:       $message          A message to be printed, logged, etc.
##               $level            The debug level of the message. If
##                                 not defined 0 will be assumed.  0 is
##                                 considered a normal message, 1 and 
##                                 higher is considered a debug message.
##  
##  Output:      Prints to STDOUT, to LOGFILE, both, or none depending 
##               on the state of the program.
##  
##  Example:     ("WARNING: We believe in generic error messages... NOT!\n", 1);
###############################################################################################
sub printmsg {
  my %incoming = ();
  (
    $incoming{'message'},
    $incoming{'level'}
  ) = @_;
  $incoming{'level'} = 0 if (!defined($incoming{'level'}));
  
  ## Print to the log file if $conf{'logging'} is true and $conf{'debug'} >= $incoming{'level'}
  if ($conf{'logging'} and ($conf{'debug'} >= $incoming{'level'}) ) {
      print LOGFILE localtime() . " - $incoming{'message'}\n";
  }
  
  ## Print to STDOUT always if debugging is enabled, or if the mode is NOT "running".
  if ( ( ($conf{'debug'} >= 1) or ($conf{'mode'} ne "running") ) and ($conf{'debug'} >= $incoming{'level'}) ) {
      print localtime() . " - $incoming{'message'}\n";
  }
  
  ## Return
  return(1);
}












###############################################################################################
## FUNCTION:    
##   openLogFile ( $filename )
## 
## 
## DESCRIPTION: 
##   Opens the file $filename and attaches it to the filehandle "LOGFILE".  Returns 0 on success
##   and non-zero on failure.  Error codes are listed below, and the error message gets set in
##   global variable $!.
##   
##   
## Example: 
##   openFile ("/var/log/scanAlert.log");
##
###############################################################################################
sub openLogFile {
    ## Get the incoming filename
    my $filename = $_[0];
    
    ## Make sure our file exists, and if the file doesn't exist then create it
    if ( ! -f $filename ) {
        printmsg("NOTICE: The file [$filename] does not exist.  Creating it now with mode [0600].", 0);
        open (LOGFILE, ">>$filename");
        close LOGFILE;
        chmod (0600, $filename);
    }
    
    ## Now open the file and attach it to a filehandle
    open (LOGFILE,">>$filename") or return (1);
    
    ## Put the file into non-buffering mode
    select LOGFILE;
    $| = 1;
    select STDOUT;
    
    ## Tell the rest of the program that we can log now
    $conf{'logging'} = "yes";
    
    ## Return success
    return(0);
}












###############################################################################################
## FUNCTION:    
##   openFIFO ( $filename )
## 
## 
## DESCRIPTION: 
##   Opens the fifo $filename for reading and attaches it to the filehandle "FIFO".  Returns 0 on success
##   and non-zero on failure.  Error codes are listed below, and the error message gets set in
##   global variable $!. 
##   
##
## Example: 
##   openFIFO ("/var/log/fifo.scanAlert");
##
###############################################################################################
sub openFIFO {
    ## Get the incoming filename
    my $filename = $_[0];
    
    ## Make sure our fifo exists, and if the fifo doesn't exist then create it
    if ( ! -p $filename ) {
        quit("ERROR: FIFO does not exist or is not a FIFO at [$conf{'fifo'}]\nDid you forget to create it and configure syslogd to log to it?\n", 1);
    }
    
    ## Open the FIFO socket we're supposed to listen to and attach it to a filehandle
    open (FIFO,"< $filename") or return(1);
    
    ## Return success
    return(0);
}











###############################################################################################
## FUNCTION:    
##   syslog (string $socketName, int $priority, int $facility, string $message)
## 
## 
## DESCRIPTION: 
##   Connects to the socket $socketName, and sends it a syslog formatted message.  
##   i.e. it syslog's $message with the priority and facility specified.
##
## Example: 
##   syslog("/dev/log", 6, 1, "Test syslog message");
##
##
## Priorities (on a Linux system)
##   LOG_EMERG       0       system is unusable
##   LOG_ALERT       1       action must be taken immediately
##   LOG_CRIT        2       critical conditions
##   LOG_ERR         3       error conditions
##   LOG_WARNING     4       warning conditions
##   LOG_NOTICE      5       normal but significant condition
##   LOG_INFO        6       informational
##   LOG_DEBUG       7       debug-level messages
##
##
## Facilities (on a Linux system)
##   LOG_KERN        0       kernel messages
##   LOG_USER        1       random user-level messages
##   LOG_MAIL        2       mail system
##   LOG_DAEMON      3       system daemons
##   LOG_AUTH        4       security/authorization messages
##   LOG_SYSLOG      5       messages generated internally by syslogd
##   LOG_LPR         6       line printer subsystem
##   LOG_NEWS        7       network news subsystem
##   LOG_UUCP        8       UUCP subsystem
##   LOG_CRON        9       clock daemon
##   LOG_AUTHPRIV    10      security/authorization messages (private)
##   LOG_FTP         11      ftp daemon
##
###############################################################################################
sub syslog {
  
  ## Get incoming variables
  my %incoming = ();
  (
    $incoming{'socketName'},
    $incoming{'priority'},
    $incoming{'facility'},
    $incoming{'message'},
  ) = @_;
  
  ## Set defaults if some variables are not defined
  $incoming{'socketName'}  = "/dev/log"         if (!defined($incoming{'socketName'}));
  $incoming{'priority'}    = 6                  if (!defined($incoming{'priority'}));
  $incoming{'facility'}    = 1                  if (!defined($incoming{'facility'}));
  $incoming{'message'}     = "Default Message"  if (!defined($incoming{'message'}));
  my $return = 1;

  ## Make sure values are sane (black magic)
  $incoming{'priority'} = $incoming{'priority'} & 0x07;
  $incoming{'facility'} = $incoming{'facility'} & 0x3f8;
  
  
  ## Debuging
  printmsg("syslog(): Syslogging the message [$incoming{'message'}]", 1);
  
  ## Open a UNIX socket in dgram mode with udp protocol. 
  socket(SOCKET, PF_UNIX, SOCK_DGRAM, 0) || ($return = 0);
  if ($return == 0) { printmsg("$$ - syslog() - Error: $!", 0); }
  
  ## Connect our socket to SOCKET
  connect(SOCKET, sockaddr_un ($incoming{'socketName'})) || ($return = 0);
  if ($return == 0) { printmsg("$$ - syslog() - Error: $!", 0); }
  
  
  
  ## Generate syslog value for the priority and facility (black magic)
  $incoming{'value'} = ($incoming{'priority'} + (8 * $incoming{'facility'}));

  ## Sending the message to the syslog socket
  print SOCKET "<$incoming{'value'}>$incoming{'message'}" || ($return = 0);
  if ($return == 0) { printmsg("$$ - syslog() - Error: $!", 0); }
  
  
  ## Close the socket
  close SOCKET;
  
  
  ## Return success
  return($return);
}











######################################################################
##  Function:    quit (string $message, int $errorLevel)
##  
##  Description: Exits the program, optionally printing $message.  It 
##               returns an exit error level of $errorLevel to the 
##               system  (0 means no errors, and is assumed if empty.)
##
##  Example:     quit("Exiting program normally", 0);
######################################################################
sub quit {
  my %incoming = ();
  (
    $incoming{'message'},
    $incoming{'errorLevel'}
  ) = @_;
  $incoming{'errorLevel'} = 0 if (!defined($incoming{'errorLevel'}));
  
  ## Print exit message
  if ($incoming{'message'}) { 
      printmsg($incoming{'message'}, 0);
      if ($incoming{'errorLevel'} >= 0) {
          syslog("/dev/log", 6, 1, $incoming{'message'});
      }
  }
  
  ## Exit
  exit($incoming{'errorLevel'});
}






